package de.opalpiraten.scenepainter.gui.render;
import java.awt.*;
import java.awt.image.*;
import java.util.*;
import java.awt.geom.*;

import com.samskivert.swing.Label;
import com.samskivert.swing.MultiLineLabel;

/**
 * The TextSceneElement is able to render various kinds of Texts. 
 */
public class TextSceneElement extends SceneElement {
  
  /**
   * For calculating the bounds without having to create a new Graphics
   * object each time.
   */
  private static final Graphics2D context= new BufferedImage(1,1,BufferedImage.TYPE_INT_ARGB).createGraphics();
  
  /**
   * Supported Textstyle
   */
  public static final String OUTLINE = "outline";
  
  /**
   * Supported Textstyle
   */
  public static final String BOX = "box";
  
  /**
   * Supported TextStyle
   */
  public static final String SPEECH ="speech";
  
  /**
   * Supported Textstyle
   */
  public static final String PLAIN = "plain";
  
  /**
   * Supported Textstyle
   */
  public static final String THOUGHT = "thought";
  
  /**
   * Supported Textstyle
   */
  public static final String EMOTION = "emotion";
  
  /**
   * Dock the tail bottom left
   */
  public static final int BOTTOMLEFT=0;
  
  /**
   * Dock the tail middle bottom
   */
  public static final int BOTTOM=1;
  
  /**
   * Dock the tail bottom right
   */
  public static final int BOTTOMRIGHT=2;
  
  /**
   * Dock the tail to the right (unsupported)
   */
  public static final int RIGHT=3;
  
  /**
   * Dock the tail top right
   */
  public static final int TOPRIGHT=4;
  
  /**
   * Dock the tail top center
   */
  public static final int TOP = 5;
  
  /**
   * Dock the tail top left
   */
  public static final int TOPLEFT = 6;
  
  /**
   * Dock the tail left (unsupported)
   */
  public static final int LEFT = 7;
  
  /**
   * where to dock the tail
   */
  private int taildock;

  /**
   * The Font to use for rendering
   */
  private Font font;
  
  /**
   * The actual text to draw
   */
  private String text;
  
  /**
   * The color of the text
   */
  private Color textcolor;
  
  /**
   * The color of the border of the background shape (may be null).
   */
  private Color bordercolor;
  
  /**
   * The color to fill the background with (may be null)
   */
  private Color backcolor;
  
  /**
   * The textstyle to use
   */
  private String style;
  
  /**
   * Length of the tail (for speech or thought bubbles)
   */
  private int taillength=25;
  
   
  public TextSceneElement() {}
  
  
  /**
   * Set the Font to render with
   */
  public void setFont(Font f) { font =f; cache=null;}
  
  /**
   * Query the Font for rendering
   */
  public Font getFont() { return font;}
  
  /**
   * Set the Text to render
   */
  public void setText(String t) { 
    text=t; 
    cache=null;
    int len = t.length();
    if (len>14) setName("\""+t.substring(0,14)+"...\"");
    else setName("\""+t.substring(0,len)+"\"");
  }
  
  /**
   * Query text to render
   */
  public String getText() { return text;}
  
  /**
   * Set color of the text
   */
  public void setTextcolor(Color c) { textcolor=c; cache=null;}
  
  /**
   * Query color of the text
   */
  public Color getTextcolor() { return textcolor; }
  
  /**
   * Set color of the border
   */
  public void setBordercolor(Color c) {bordercolor =c; cache=null;}
  
  /**
   * Query color of the border
   */
  public Color getBordercolor() { return bordercolor;}
  
  /**
   * Set color of the backgroundshape
   */
  public void setBackcolor(Color c) { backcolor=c; cache=null;}
  
  /**
   * Query the color of the backgroundshape
   */
  public Color getBackcolor() { return backcolor;}
  
  /**
   * Set the style of this TextElement
   */
  public void setStyle(String s) {
    style = s;
    cache=null;
  }
  
  /**
   * Query the style of this element
   */
  public String getStyle() {
    return style;
  }
  
  /**
   * Set the length of the tail (for speech bubbles)
   */
  public void setTaillength(int l) {
    taillength=l;
  }
  
  /**
   * Query the length of the tail
   */
  public int getTaillength() {
    return taillength;
  }
  
  /**
   * Where to put the tail
   */
  public void setTaildock(int i) {
    taildock=i;
  }
  
  /**
   * Query dock position
   */
  public int getTaildock() {
    return taildock;
  }
  
  
  // Overridden method
  public void scrollView(boolean b) {
    if (b) {
      taillength++;
    }
    else {
      taillength--;
      if (taillength<=0) taillength=0;
    }
    
    cache=null;
  }

  // Implementation of abstract method
  public BufferedImage buildCache() {
    BufferedImage draft =null;
    try {
      
      if (style.equals(OUTLINE)) draft=buildOutline();
      if (style.equals(BOX)) draft=buildBox();
      if (style.equals(SPEECH)) draft=buildSpeech();
      if (style.equals(THOUGHT)) draft=buildThought();
      if (style.equals(EMOTION)) draft=buildEmote();
      if (style.equals(PLAIN)) draft=drawText();
      
    }
    catch (Exception e) {
      e.printStackTrace();
      draft=new BufferedImage(50,50,BufferedImage.TYPE_INT_ARGB);
      Graphics2D g = draft.createGraphics();
      g.setColor(Color.RED);
      g.fillOval(0,0,50,50);
      g.setColor(Color.white);
      g.fillRect(5,20,40,10);
    }
    
    BufferedImage ret = new BufferedImage(draft.getWidth(),draft.getHeight(),BufferedImage.TYPE_INT_ARGB);
    Graphics2D g = (Graphics2D)ret.createGraphics();
    g.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER,alpha));
    g.drawImage(draft,0,0,null);
    return ret;
  }
  
  /**
   * Emotional text
   */
  private BufferedImage buildEmote() {
    BufferedImage txt = drawText();
    int w=txt.getWidth()+2*taillength; 
    int h=txt.getHeight()+2*taillength;
    BufferedImage draft = new BufferedImage(w+1,h+1,BufferedImage.TYPE_INT_ARGB);
    Graphics2D gr = draft.createGraphics();
    gr.setRenderingHint(RenderingHints.KEY_ANTIALIASING,RenderingHints.VALUE_ANTIALIAS_ON);
    
    GeneralPath path = new GeneralPath();
    path.moveTo(0,0);
    path.curveTo(0,0, taillength,(h/2) ,0,h);
    path.curveTo(0,h, w/2,h-taillength ,w,h);
    path.curveTo(w,h,w-taillength,h/2,w,0);
    path.curveTo(w,0,w/2,taillength,0,0);
    gr.setColor(backcolor);
    gr.fill(path);
    gr.setColor(bordercolor);
    gr.draw(path);
    gr.setColor(textcolor);
    gr.drawImage(txt,taillength,taillength,null);

    return draft;
  }
  
  /**
   * Outline Text.
   */
  private BufferedImage buildOutline() {
    Label lbl = new Label(text,Label.OUTLINE,textcolor,bordercolor,font);
    lbl.layout(context);
    Dimension d = lbl.getSize();
    BufferedImage draft=new BufferedImage(d.width,d.height,BufferedImage.TYPE_INT_ARGB);
    lbl.render(draft.createGraphics(),0,0); 
    return draft;
  }
  
  /**
   * Build a boxed text
   */
  private BufferedImage buildBox() {
    BufferedImage txt = drawText();
    int w=txt.getWidth()+6;
    int h=txt.getHeight()+6;
    BufferedImage draft = new BufferedImage(w,h,BufferedImage.TYPE_INT_ARGB);
    Graphics2D gr = draft.createGraphics();
    gr.setColor(backcolor);
    gr.fillRect(0,0,w,h);
    gr.setColor(bordercolor);
    gr.drawRect(0,0,w-1,h-1);
    gr.drawImage(txt,null,3,2);
    return draft;
  }
  
  /**
   * Build a speechbubble
   */
  private BufferedImage buildSpeech() {
    BufferedImage txt = drawText();
    Dimension d=new Dimension(txt.getWidth(),txt.getHeight());
    
    int spacing=12;
    
    BufferedImage draft = new BufferedImage(d.width+2*spacing+1,d.height+2*spacing+taillength,BufferedImage.TYPE_INT_ARGB);
    Graphics2D g = draft.createGraphics();
    g.setRenderingHint(RenderingHints.KEY_ANTIALIASING,RenderingHints.VALUE_ANTIALIAS_ON);
    
    int dia = (int) (Math.sqrt(spacing*spacing + spacing*spacing) * 2);
    int xpoints[] = {0,2*spacing,2*spacing+10};
    int ypoints[] = {draft.getHeight(),d.height+2*spacing-1,d.height+2*spacing-1};
    
    switch(taildock) {
      case (BOTTOMLEFT): {
        g.setColor(backcolor);
        g.fillRoundRect(0,0,d.width+2*spacing,d.height+2*spacing,dia,dia);
        g.setColor(bordercolor);
        g.drawRoundRect(0,0,d.width+2*spacing,d.height+2*spacing,dia,dia);
        g.setColor(backcolor);
      
        g.fillPolygon(xpoints,ypoints,3);
        g.setColor(bordercolor);
        g.drawLine(xpoints[0],ypoints[0],xpoints[1]-1,ypoints[2]+1);
        g.drawLine(xpoints[0],ypoints[0],xpoints[2]-1,ypoints[2]+1);
        g.drawImage(txt,spacing,spacing,null);
        break;
      }
    
      case(BOTTOM): {
        xpoints[0] = (int)draft.getWidth()/2;
        xpoints[1] = (int)draft.getWidth()/2-5; //2*spacing;
        xpoints[2] = (int)draft.getWidth()/2+5; //2*spacing-10;
      
        g.setColor(backcolor);
        g.fillRoundRect(0,0,d.width+2*spacing,d.height+2*spacing,dia,dia);
        g.setColor(bordercolor);
        g.drawRoundRect(0,0,d.width+2*spacing,d.height+2*spacing,dia,dia);
        g.setColor(backcolor);
      
        g.fillPolygon(xpoints,ypoints,3);
        g.setColor(bordercolor);
        g.drawLine(xpoints[0],ypoints[0],xpoints[1],ypoints[2]+1);
        g.drawLine(xpoints[0],ypoints[0],xpoints[2],ypoints[2]+1);
        g.drawImage(txt,spacing,spacing,null);
        break;
      }
    
      case(BOTTOMRIGHT): {
        xpoints[0] = draft.getWidth();
        xpoints[1] = draft.getWidth()-2*spacing;
        xpoints[2] = draft.getWidth()-2*spacing-10;
      
        g.setColor(backcolor);
        g.fillRoundRect(0,0,d.width+2*spacing,d.height+2*spacing,dia,dia);
        g.setColor(bordercolor);
        g.drawRoundRect(0,0,d.width+2*spacing,d.height+2*spacing,dia,dia);
        g.setColor(backcolor);
      
        g.fillPolygon(xpoints,ypoints,3);
        g.setColor(bordercolor);
        g.drawLine(xpoints[0],ypoints[0],xpoints[1],ypoints[2]+1);
        g.drawLine(xpoints[0],ypoints[0],xpoints[2],ypoints[2]+1);
        g.drawImage(txt,spacing,spacing,null);
        break;
      }

      case(RIGHT): {
        draft = new BufferedImage(d.width+2*spacing+1+taillength,d.height+2*spacing+1,BufferedImage.TYPE_INT_ARGB);
        g = draft.createGraphics();
        g.setRenderingHint(RenderingHints.KEY_ANTIALIASING,RenderingHints.VALUE_ANTIALIAS_ON);
        xpoints[0] = (int)draft.getWidth();
        xpoints[1] = (int)draft.getWidth()-taillength-1;
        xpoints[2] = (int)draft.getWidth()-taillength-1;
        ypoints[0] = (int)draft.getHeight()/2;
        ypoints[1] = (int)draft.getHeight()/2-5;
        ypoints[2] = (int)draft.getHeight()/2+5; 
      
        g.setColor(backcolor);
        g.fillRoundRect(0,0,d.width+2*spacing,d.height+2*spacing,dia,dia);
        g.setColor(bordercolor);
        g.drawRoundRect(0,0,d.width+2*spacing,d.height+2*spacing,dia,dia);
        g.setColor(backcolor);
      
        g.fillPolygon(xpoints,ypoints,3);
        g.setColor(bordercolor);
        g.drawLine(xpoints[0],ypoints[0],xpoints[1],ypoints[1]-1);
        g.drawLine(xpoints[0],ypoints[0],xpoints[2],ypoints[2]-1);
        g.drawImage(txt,spacing,spacing,null);
        break;
      }

      case(TOPRIGHT): {
        xpoints[0] = (int)draft.getWidth();
        xpoints[1] = (int)draft.getWidth()-2*spacing;
        xpoints[2] = (int)draft.getWidth()-2*spacing-10;
        ypoints[0] = 0;
        ypoints[1] = taillength;
        ypoints[2] = taillength; 
      
        g.setColor(backcolor);
        g.fillRoundRect(0,taillength-1,d.width+2*spacing,d.height+2*spacing,dia,dia);
        g.setColor(bordercolor);
        g.drawRoundRect(0,taillength-1,d.width+2*spacing,d.height+2*spacing,dia,dia);
        g.setColor(backcolor);
      
        g.fillPolygon(xpoints,ypoints,3);
        g.setColor(bordercolor);
        g.drawLine(xpoints[0],ypoints[0],xpoints[1],ypoints[2]-1);
        g.drawLine(xpoints[0],ypoints[0],xpoints[2],ypoints[2]-1);
        g.drawImage(txt,spacing,spacing+taillength-1,null);
        break;
      }

      case(TOP): {
        xpoints[0] = (int)draft.getWidth()/2;
        xpoints[1] = (int)draft.getWidth()/2+5;
        xpoints[2] = (int)draft.getWidth()/2-5;
        ypoints[0] = 0;
        ypoints[1] = taillength;
        ypoints[2] = taillength;
      
        g.setColor(backcolor);
        g.fillRoundRect(0,taillength-1,d.width+2*spacing,d.height+2*spacing,dia,dia);
        g.setColor(bordercolor);
        g.drawRoundRect(0,taillength-1,d.width+2*spacing,d.height+2*spacing,dia,dia);
        g.setColor(backcolor);
      
        g.fillPolygon(xpoints,ypoints,3);
        g.setColor(bordercolor);
        g.drawLine(xpoints[0],ypoints[0],xpoints[1],ypoints[2]-1);
        g.drawLine(xpoints[0],ypoints[0],xpoints[2],ypoints[2]-1);
        g.drawImage(txt,spacing,spacing+taillength-1,null);
        break;
      }

      case(TOPLEFT): {
        xpoints[0] = 0;
        xpoints[1] = spacing*2;
        xpoints[2] = spacing*2+10;
        ypoints[0] = 0;
        ypoints[1] = taillength;
        ypoints[2] = taillength;
      
        g.setColor(backcolor);
        g.fillRoundRect(0,taillength-1,d.width+2*spacing,d.height+2*spacing,dia,dia);
        g.setColor(bordercolor);
        g.drawRoundRect(0,taillength-1,d.width+2*spacing,d.height+2*spacing,dia,dia);
        g.setColor(backcolor);
      
        g.fillPolygon(xpoints,ypoints,3);
        g.setColor(bordercolor);
        g.drawLine(xpoints[0],ypoints[0],xpoints[1],ypoints[2]-1);
        g.drawLine(xpoints[0],ypoints[0],xpoints[2],ypoints[2]-1);
        g.drawImage(txt,spacing,spacing+taillength-1,null);
        break;
      }

      case(LEFT): {
        draft = new BufferedImage(d.width+2*spacing+1+taillength,d.height+2*spacing+1,BufferedImage.TYPE_INT_ARGB);
        g = draft.createGraphics();
        g.setRenderingHint(RenderingHints.KEY_ANTIALIASING,RenderingHints.VALUE_ANTIALIAS_ON);
        xpoints[0] = 0;
        xpoints[1] = taillength+1;
        xpoints[2] = taillength+1;
        ypoints[0] = (int)draft.getHeight()/2;
        ypoints[1] = (int)draft.getHeight()/2-5;
        ypoints[2] = (int)draft.getHeight()/2+5; 
      
        g.setColor(backcolor);
        g.fillRoundRect(taillength,0,d.width+2*spacing,d.height+2*spacing,dia,dia);
        g.setColor(bordercolor);
        g.drawRoundRect(taillength,0,d.width+2*spacing,d.height+2*spacing,dia,dia);
        g.setColor(backcolor);
      
        g.fillPolygon(xpoints,ypoints,3);
        g.setColor(bordercolor);
        g.drawLine(xpoints[0],ypoints[0],xpoints[1]-1,ypoints[1]);
        g.drawLine(xpoints[0],ypoints[0],xpoints[2]-1,ypoints[2]);
        g.drawImage(txt,spacing+taillength,spacing,null);
        break;
      }

    }
    
    return draft;
  }


  /**
   * Build a thoughtbubble
   */
  private BufferedImage buildThought() {
    BufferedImage txt = drawText();
    Dimension d=new Dimension(txt.getWidth(),txt.getHeight());
    d.width=d.width+ (10-d.width%10);
    d.height=d.height +(10-d.height%10);
    int spacing=15;
    
    BufferedImage draft = new BufferedImage(d.width+2*spacing+1,d.height+2*spacing+taillength,BufferedImage.TYPE_INT_ARGB);
    Graphics2D g = draft.createGraphics();
    g.setRenderingHint(RenderingHints.KEY_ANTIALIASING,RenderingHints.VALUE_ANTIALIAS_ON);
    
    GeneralPath path = new GeneralPath();
    path.moveTo(spacing,spacing);
    for (int i=spacing-5;i<d.height+spacing-5;i=i+10) path.curveTo(spacing-5,i,0,i+5,spacing-5,i+10);
    for (int i=spacing-5;i<d.width+spacing-1;i=i+10) path.curveTo(i,d.height+spacing+5, i+5,d.height+10+spacing+5, i+10,d.height+spacing+5);
    for (int i=d.height+spacing;i>spacing;i=i-10) path.curveTo(d.width+1*(spacing+5),i, d.width+10+1*(spacing+5),i-5,d.width+1*(spacing+5),i-10);
    for (int i=d.width+spacing;i>spacing+1;i=i-10) path.curveTo(i,spacing,i-5,0,i-10,spacing);
    
    path.closePath();
    
    g.setColor(backcolor);
    g.fill(path);
    g.setColor(bordercolor);
    g.draw(path);
    g.setColor(textcolor);
    g.drawImage(txt,spacing,spacing,null);
    
    return draft;
  }
  
  
  /**
   * Convinience method to render the text on a buffered image
   */
  private BufferedImage drawText() {
    int width=1;
    int height=0;
    BufferedImage  ret = new BufferedImage(1,1,BufferedImage.TYPE_INT_ARGB);
    Graphics2D gr = ret.createGraphics();
    gr.setFont(font);
    FontMetrics fm = gr.getFontMetrics(font);
    
    StringTokenizer st = new StringTokenizer(text,"\n");
    while(st.hasMoreTokens()) {
      Rectangle2D r = fm.getStringBounds(st.nextToken(),gr);
      if (r.getWidth()>width) width=(int)(r.getWidth()+1);
      height++;
    }
    height=height*(fm.getLeading()+fm.getMaxAscent()+fm.getMaxDescent())+fm.getMaxDescent();
    
    ret = new BufferedImage(width,height,BufferedImage.TYPE_INT_ARGB);
    gr = ret.createGraphics();
    gr.setFont(font);
    gr.setColor(textcolor);
    st = new StringTokenizer(text,"\n");
    int y=1;
    while(st.hasMoreTokens()) {
      gr.drawString(st.nextToken(),0,y*fm.getHeight());
      y++;
    }
    return ret;
  }

}
